@@ -1,7 +1,7 @@
 /*
  * cddb-slave.c: Implementation for the GNOME/Media/CDDBSlave2 interface.
  *
- * Copyright (C) 2001 Ximian, Inc.
+ * Copyright (C) 2001-2002 Iain Holmes.
  * 
  * Authors: Iain Holmes  <iain@ximian.com>
  */
@@ -23,15 +23,20 @@
 #include <bonobo/bonobo-exception.h>
 #include <gnome.h>
 
+#include <gconf/gconf-client.h>
+
 /* Use local copy of gnet */
 #include "gnet.h"
 
 #include <orbit/orbit.h>
 
 #include "cddb-slave.h"
+#include "cddb-parser.h"
 #include "GNOME_Media_CDDBSlave2.h"
 
 
+static GConfClient *client = NULL;
+
 #define PARENT_TYPE BONOBO_OBJECT_TYPE
 static BonoboObjectClass *parent_class = NULL;
 
@@ -42,6 +47,8 @@ typedef enum {
 } CDDBSlaveAccess;
 
 struct _CDDBSlavePrivate {
+	char *name;
+	char *hostname;
 	char *server; /* Server address */
 	int port; /* Server port */
 
@@ -75,6 +82,7 @@ typedef struct _ConnectionData {
 
 	GTcpSocket *socket;
 	GIOChannel *iochannel;
+	guint tag;
 	
 	ConnectionMode mode;
 	
@@ -90,6 +98,7 @@ typedef struct _ConnectionData {
 } ConnectionData;
 
 static GHashTable *pending_requests = NULL;
+static GHashTable *cddb_cache = NULL;
 
 static void do_hello (ConnectionData *cd);
 
@@ -140,6 +149,10 @@ do_goodbye_response (ConnectionData *cd,
 	}
 
 	/* Disconnect */
+	if (cd->tag) {
+		g_source_remove (cd->tag);
+	}
+	
   	gnet_tcp_socket_unref (cd->socket);
   	g_io_channel_unref (cd->iochannel);
 	
@@ -167,6 +180,28 @@ do_goodbye_response (ConnectionData *cd,
 	return FALSE;
 }
 
+static void
+clear_entry_from_cache (const char *discid)
+{
+	CDDBEntry *entry;
+	int i;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	/* Set the fields of the entry to unknown */
+	entry->revision = 0;
+
+	g_hash_table_insert (entry->fields, g_strdup ("DTITLE"),
+			     g_string_new (_("Unknown / Unknown")));
+	for (i = 0; i < entry->ntrks; i++) {
+		char *name;
+		GString *title;
+
+		name = g_strdup_printf ("TTITLE%d", i);
+		title = g_string_new (_("Unknown"));
+		g_hash_table_insert (entry->fields, name, title);
+	}
+}
+
 static void
 do_goodbye (ConnectionData *cd)
 {
@@ -187,6 +222,7 @@ do_read_response (ConnectionData *cd,
 	CDDBSlave *cddb = cd->cddb;
 	int code;
 	gboolean more = FALSE;
+	gboolean disconnect = FALSE;
 	static gboolean waiting_for_terminator = FALSE;
 	static FILE *handle = NULL;
 	
@@ -216,20 +252,33 @@ do_read_response (ConnectionData *cd,
 			}
 
 			waiting_for_terminator = TRUE;
+		} else {
+
+			g_assert (handle != NULL);
+		
+			/* Write the line */
+			fputs (response, handle);
 		}
 
-		g_assert (handle != NULL);
-		
-		/* Write the line */
-		fputs (response, handle);
-
 		if (response[0] == '.') {
+			CDDBEntry *entry;
+			char *filename, *dirname;
+			
 			/* Found terminator */
 			fclose (handle);
 
+			dirname = gnome_util_prepend_user_home (".cddbslave");
+			filename = g_concat_dir_and_file (dirname, cd->discid);
+			g_free (dirname);
+			
+			entry = g_hash_table_lookup (cddb_cache, cd->discid);
+			cddb_entry_parse_file (entry, filename);
+			g_free (filename);
+			
 			/* Reset the static variables for next time */
 			handle = NULL;
 			more = FALSE;
+			disconnect = TRUE;
 			waiting_for_terminator = FALSE;
 		} else {
 			more = TRUE;
@@ -241,18 +290,24 @@ do_read_response (ConnectionData *cd,
 		g_print ("Specified CDDB entry not found\n");
 		g_print ("%s\n", response);
 		more = FALSE;
+		disconnect = TRUE;
+		clear_entry_from_cache (cd->discid);
 		break;
 
 	case 402:
 		g_print ("Server error\n");
 		g_print ("%s\n", response);
 		more = FALSE;
+		disconnect = TRUE;
+		clear_entry_from_cache (cd->discid);
 		break;
 
 	case 403:
 		g_print ("Database entry is corrupt\n");
 		g_print ("%s\n", response);
 		more = FALSE;
+		disconnect = TRUE;
+		clear_entry_from_cache (cd->discid);
 		break;
 
 	case 409:
@@ -266,7 +321,7 @@ do_read_response (ConnectionData *cd,
 
 	}
 
-	if (more == FALSE) {
+	if (disconnect == TRUE) {
 		do_goodbye (cd);
 	}
 		
@@ -456,6 +511,7 @@ do_query_response (ConnectionData *cd,
 				} else {
 				/* Need to disconnect here...
 				   none of our matches matched */
+					clear_entry_from_cache (cd->discid);
 					do_goodbye (cd);
 				}
 			}
@@ -491,12 +547,14 @@ do_query_response (ConnectionData *cd,
 		g_print ("No match found\n");
 		g_print ("%s\n", response);
 		more = FALSE;
+		disconnect = TRUE;
 		break;
 
 	case 403:
 		g_print ("Database entry is corrupt\n");
 		g_print ("%s\n", response);
 		more = FALSE;
+		disconnect = TRUE;
 		break;
 
 	case 409:
@@ -510,6 +568,7 @@ do_query_response (ConnectionData *cd,
 	default:
 		g_print ("Unknown response\n");
 		g_print ("%s\n", response);
+		disconnect = TRUE;
 		more = FALSE;
 		break;
 	}
@@ -521,6 +580,11 @@ do_query_response (ConnectionData *cd,
 		do_read (cd, cat, discid, dtitle);
 	}
 
+	if (disconnect == TRUE) {
+		clear_entry_from_cache (cd->discid);
+		do_goodbye (cd);
+	}
+	
 	g_free (cat);
 	g_free (discid);
 	g_free (dtitle);
@@ -564,6 +628,8 @@ do_hello_response (ConnectionData *cd,
 		g_print ("%s\n", response);
 
 		/* Disconnect here */
+		clear_entry_from_cache (cd->discid);
+		do_goodbye (cd);
 		break;
 
 	case 402:
@@ -594,8 +660,11 @@ do_hello (ConnectionData *cd)
 	/* Send the Hello command
 	   CDDB howto says these shouldn't be hardcoded,
 	   but that seems to be a privacy issue */
-	hello = g_strconcat ("cddb hello johnsmith 198.172.174.22 ",
-			     cd->name, " ", cd->version, "\n", NULL);
+	hello = g_strdup_printf ("cddb hello %s %s %s %s\n",
+				 cd->cddb->priv->name,
+				 cd->cddb->priv->hostname,
+				 cd->name,
+				 cd->version);
 	
 	/* Need to check the return of this one */
 	status = gnet_io_channel_writen (cd->iochannel, hello,
@@ -658,6 +727,7 @@ do_open_response (ConnectionData *cd,
 	} else {
 		/* Do something to indicate that we can't contact server */
 		/* Close connection to tell listeners we're not doing anything */
+		clear_entry_from_cache (cd->discid);
 		do_goodbye (cd);
 	}
 
@@ -742,6 +812,7 @@ read_from_server (GIOChannel *iochannel,
 	return TRUE;
 
  error:
+	clear_entry_from_cache (cd->discid);
 	return FALSE;
 }
 
@@ -756,19 +827,20 @@ open_cb (GTcpSocket *sock,
 
 	cd->socket = sock;
 	if (status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK) {
-		g_warning ("Error opening %s:%s",
+		g_warning ("Error opening %s:%d",
 			   cd->cddb->priv->server,
 			   cd->cddb->priv->port);
 
 		/* notify listeners */
+		clear_entry_from_cache (cd->discid);
 		return;
 	}
 
 	sin = gnet_tcp_socket_get_iochannel (sock);
 	cd->iochannel = sin;
 	
-	g_io_add_watch (sin, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-			read_from_server, data);
+	cd->tag = g_io_add_watch (sin, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+				  read_from_server, data);
 }
 
 static void
@@ -791,27 +863,23 @@ cddb_send_cmd (ConnectionData *data)
 static gboolean
 cddb_check_cache (const char *discid)
 {
-	char *dirname, *discname;
+	if (g_hash_table_lookup (cddb_cache, discid) == NULL) {
+		char *filename;
+		CDDBEntry *entry;
 
-	g_return_val_if_fail (discid != NULL, FALSE);
+		filename = g_build_filename (g_get_home_dir (),
+					     ".cddbslave",
+					     discid, NULL);
+		entry = cddb_entry_new_from_file (filename);
+		g_free (filename);
+		if (entry == NULL) {
+			return FALSE;
+		}
 
-	dirname = gnome_util_prepend_user_home (".cddbslave");
-	if (g_file_test (dirname, G_FILE_TEST_IS_DIR) == FALSE) {
-		/* Cache dir doesn't exist */
-		g_free (dirname);
-
-		mkdir (dirname, S_IRWXU);
-		return FALSE;
-	}
-
-	discname = g_concat_dir_and_file (dirname, discid);
-	g_free (dirname);
-	if (g_file_test (discname, G_FILE_TEST_IS_REGULAR) == TRUE) {
-		g_free (discname);
+		g_hash_table_insert (cddb_cache, g_strdup (discid), entry);
 		return TRUE;
 	} else {
-		g_free (discname);
-		return FALSE;
+		return TRUE;
 	}
 }
 
@@ -834,23 +902,28 @@ impl_GNOME_Media_CDDBSlave2_query (PortableServer_Servant servant,
 				   CORBA_Environment *ev)
 {
 	CDDBSlave *cddb;
+	CDDBEntry *entry;
 	ConnectionData *cd;
 	char *request, *safe_offsets;
 	char *username, *hostname, *fullname, *uri;
 
 	cddb = cddb_slave_from_servant (servant);
 
-	g_warning ("Request: %s", discid);
-	/* Do stuff */
 	if (cddb_check_cache (discid) == TRUE) {
-		g_print ("In cache\n");
-		cddb_slave_notify_listeners (cddb, discid, GNOME_Media_CDDBSlave2_OK);
+		cddb_slave_notify_listeners (cddb, discid,
+					     GNOME_Media_CDDBSlave2_OK);
 		return;
 	}
 
+	/* Make an entry */
+	entry = cddb_entry_new (discid, ntrks, offsets, nsecs);
+	g_hash_table_insert (cddb_cache, g_strdup (discid), entry);
+	
 	cd = g_new (ConnectionData, 1);
 	cd->cddb = cddb;
 	g_object_ref (cddb);
+
+	cd->tag = 0;
 	
 	cd->name = g_strdup_printf ("%s(CDDBSlave2)", name);
 	cd->version = g_strdup (version);
@@ -864,6 +937,450 @@ impl_GNOME_Media_CDDBSlave2_query (PortableServer_Servant servant,
 	cddb_send_cmd (cd);
 }
 
+static void
+impl_GNOME_Media_CDDBSlave2_save (PortableServer_Servant servant,
+				  const CORBA_char *discid,
+				  CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	gboolean write_ret;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	write_ret = cddb_entry_write_to_file (entry);
+	if (write_ret == FALSE) {
+		/* Exception */
+	}
+}
+
+static CORBA_char *
+impl_GNOME_Media_CDDBSlave2_getArtist (PortableServer_Servant servant,
+				       const CORBA_char *discid,
+				       CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	char *split, *artist;
+	GString *dtitle;
+	CORBA_char *ret;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return CORBA_string_dup ("Unknown artist");
+	}
+
+	dtitle = g_hash_table_lookup (entry->fields, "DTITLE");
+	if (dtitle == NULL) {
+		g_warning ("No field DTITLE %s", discid);
+		return CORBA_string_dup ("Unknown artist");
+	}
+
+	split = strstr (dtitle->str, " / ");
+	if (split == NULL) {
+		ret = CORBA_string_dup (dtitle->str);
+	} else {
+		artist = g_strndup (dtitle->str, split - dtitle->str);
+		ret = CORBA_string_dup (artist);
+		g_free (artist);
+	}
+	
+	return ret;
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_setArtist (PortableServer_Servant servant,
+				       const CORBA_char *discid,
+				       const CORBA_char *artist,
+				       CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dtitle;
+	char **split;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	dtitle = g_hash_table_lookup (entry->fields, "DTITLE");
+	if (dtitle == NULL) {
+		/* Should a DTITLE be created? */
+		return;
+	}
+
+	split = g_strsplit (dtitle->str, " / ", 2);
+	if (split == NULL) {
+		g_string_printf (dtitle, "%s", artist);
+	} else {
+		g_string_printf (dtitle, "%s / %s", artist, split[1]);
+		g_strfreev (split);
+	}
+}
+		
+static CORBA_char *
+impl_GNOME_Media_CDDBSlave2_getDiscTitle (PortableServer_Servant servant,
+					  const CORBA_char *discid,
+					  CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	char *split;
+	GString *dtitle;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return CORBA_string_dup ("Unknown disc");
+	}
+
+	dtitle = g_hash_table_lookup (entry->fields, "DTITLE");
+	if (dtitle == NULL) {
+		g_warning ("Entry has no DTITLE %s", discid);
+		return CORBA_string_dup ("Unknown disc");
+	}
+
+	/* FIXME: use g_strsplit */
+	split = strstr (dtitle->str, " / ");
+	if (split == NULL) {
+		return CORBA_string_dup ("");
+	} else {
+		return CORBA_string_dup (split + 3);
+	}
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_setDiscTitle (PortableServer_Servant servant,
+					  const CORBA_char *discid,
+					  const CORBA_char *title,
+					  CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	char **split;
+	GString *dtitle;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	dtitle = g_hash_table_lookup (entry->fields, "DTITLE");
+	if (dtitle == NULL) {
+		return;
+	}
+
+	split = g_strsplit (dtitle->str, " / ", 2);
+	if (split == NULL) {
+		g_string_printf (dtitle, " / %s", title);
+	} else {
+		g_string_printf (dtitle, "%s / %s", split[0], title);
+		g_strfreev (split);
+	}
+}
+
+static CORBA_short
+impl_GNOME_Media_CDDBSlave2_getNTrks (PortableServer_Servant servant,
+				      const CORBA_char *discid,
+				      CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return -1;
+	}
+
+	return entry->ntrks;
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_getAllTracks (PortableServer_Servant servant,
+					  const CORBA_char *discid,
+					  GNOME_Media_CDDBSlave2_TrackList **names_list,
+					  CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	int ntrk;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return;
+	}
+
+	*names_list = GNOME_Media_CDDBSlave2_TrackList__alloc ();
+	(*names_list)->_length = 0;
+	(*names_list)->_maximum = entry->ntrks + 1;
+	(*names_list)->_buffer = CORBA_sequence_GNOME_Media_CDDBSlave2_TrackInfo_allocbuf (entry->ntrks);
+
+	if (entry->ntrks == 0) {
+		g_warning ("No tracks\n");
+		return;
+	}
+
+	for (ntrk = 0; ntrk < entry->ntrks; ntrk++) {
+		char *name;
+		GString *result;
+
+		name = g_strdup_printf ("TTITLE%d", ntrk);
+		result = g_hash_table_lookup (entry->fields, name);
+		if (result != NULL) {
+			(*names_list)->_buffer[ntrk].name = CORBA_string_dup (result->str ? result->str : _("Unknown track"));
+		} else {
+			(*names_list)->_buffer[ntrk].name = CORBA_string_dup (_("Unknown track"));
+		}
+		g_free (name);
+
+		name = g_strdup_printf ("EXTT%d", ntrk);
+		result = g_hash_table_lookup (entry->fields, name);
+		if (result != NULL) {
+			(*names_list)->_buffer[ntrk].comment = CORBA_string_dup (result->str ? result->str : "");
+		} else {
+			(*names_list)->_buffer[ntrk].comment = CORBA_string_dup ("");
+		}
+
+  		(*names_list)->_buffer[ntrk].length = entry->lengths[ntrk];
+		
+		(*names_list)->_length++;
+
+		g_free (name);
+	}
+
+	return;
+}
+
+static void
+set_new_string (CDDBEntry *entry,
+		const char *key,
+		const char *new_contents)
+{
+	GString *dentry;
+
+	dentry = g_hash_table_lookup (entry->fields, key);
+	if (dentry == NULL) {
+		dentry = g_string_new (new_contents);
+		g_hash_table_insert (entry->fields, g_strdup (key), dentry);
+	} else {
+		g_string_printf (dentry, "%s", new_contents);
+	}
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_setAllTracks (PortableServer_Servant servant,
+					  const CORBA_char *discid,
+					  const GNOME_Media_CDDBSlave2_TrackList *list,
+					  CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	int ntrks, i;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	ntrks = list->_length;
+	for (i = 0; i < ntrks; i++) {
+		char *name, *comment, *key;
+		GString *dentry;
+		
+		name = list->_buffer[i].name;
+		comment = list->_buffer[i].comment;
+
+		if (name != NULL && *name != 0) {
+			key = g_strdup_printf ("TTITLE%d", i);
+			set_new_string (entry, key, name);
+			g_free (key);
+		}
+
+		if (comment != NULL && *name != 0) {
+			key = g_strdup_printf ("EXTT%d", i);
+			set_new_string (entry, key, comment);
+			g_free (key);
+		}
+	}
+}
+
+static CORBA_char *
+impl_GNOME_Media_CDDBSlave2_getComment (PortableServer_Servant servant,
+					const CORBA_char *discid,
+					CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dcomment;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry - %s", discid);
+		return CORBA_string_dup ("");
+	}
+
+	dcomment = g_hash_table_lookup (entry->fields, "EXTD");
+	if (dcomment == NULL) {
+		return CORBA_string_dup ("");
+	} else {
+		return CORBA_string_dup (dcomment->str ? dcomment->str : "");
+	}
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_setComment (PortableServer_Servant servant,
+					const CORBA_char *discid,
+					const CORBA_char *comment,
+					CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dcomment;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	dcomment = g_hash_table_lookup (entry->fields, "EXTD");
+	if (dcomment == NULL) {
+		dcomment = g_string_new (comment);
+		g_hash_table_insert (entry->fields, g_strdup ("EXTD"), dcomment);
+	} else {
+		g_string_printf (dcomment, "%s", comment);
+	}
+}
+
+static CORBA_short
+impl_GNOME_Media_CDDBSlave2_getYear (PortableServer_Servant servant,
+				     const CORBA_char *discid,
+				     CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dyear;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return -1;
+	}
+	
+	dyear = g_hash_table_lookup (entry->fields, "DYEAR");
+	if (dyear == NULL) {
+		return -1;
+	} else {
+		return atoi (dyear->str);
+	}
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_setYear (PortableServer_Servant servant,
+				     const CORBA_char *discid,
+				     CORBA_short year,
+				     CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dyear;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	dyear = g_hash_table_lookup (entry->fields, "DYEAR");
+	if (dyear == NULL) {
+		dyear = g_string_new ("");
+		g_string_printf (dyear, "%d", year);
+		g_hash_table_insert (entry->fields, g_strdup ("DYEAR"), dyear);
+	} else {
+		g_string_printf (dyear, "%d", year);
+	}
+}
+
+static CORBA_char *
+impl_GNOME_Media_CDDBSlave2_getGenre (PortableServer_Servant servant,
+				      const CORBA_char *discid,
+				      CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dgenre;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return CORBA_string_dup ("");
+	}
+	
+	dgenre = g_hash_table_lookup (entry->fields, "DGENRE");
+	if (dgenre == NULL) {
+		return CORBA_string_dup ("");
+	} else {
+		return CORBA_string_dup (dgenre->str ? dgenre->str : "");
+	}
+}
+
+static void
+impl_GNOME_Media_CDDBSlave2_setGenre (PortableServer_Servant servant,
+				      const CORBA_char *discid,
+				      const CORBA_char *genre,
+				      CORBA_Environment *ev)
+{
+	CDDBEntry *entry;
+	GString *dgenre;
+
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	dgenre = g_hash_table_lookup (entry->fields, "DGENRE");
+	if (dgenre == NULL) {
+		dgenre = g_string_new (genre);
+		g_hash_table_insert (entry->fields, g_strdup ("DGENRE"), dgenre);
+	} else {
+		g_string_printf (dgenre, "%s", genre);
+	}
+}
+
 
 static void
 finalize (GObject *object)
@@ -878,7 +1395,9 @@ finalize (GObject *object)
 		return;
 
 	g_free (priv->server);
-
+	g_free (priv->name);
+	g_free (priv->hostname);
+	
 	g_free (priv);
 	cddb->priv = NULL;
 
@@ -896,6 +1415,20 @@ cddb_slave_class_init (CDDBSlaveClass *klass)
 
 	parent_class = g_type_class_peek_parent (klass);
 	epv->query = impl_GNOME_Media_CDDBSlave2_query;
+	epv->save = impl_GNOME_Media_CDDBSlave2_save;
+	epv->getArtist = impl_GNOME_Media_CDDBSlave2_getArtist;
+	epv->setArtist = impl_GNOME_Media_CDDBSlave2_setArtist;
+	epv->getDiscTitle = impl_GNOME_Media_CDDBSlave2_getDiscTitle;
+	epv->setDiscTitle = impl_GNOME_Media_CDDBSlave2_setDiscTitle;
+	epv->getNTrks = impl_GNOME_Media_CDDBSlave2_getNTrks;
+	epv->getAllTracks = impl_GNOME_Media_CDDBSlave2_getAllTracks;
+	epv->setAllTracks = impl_GNOME_Media_CDDBSlave2_setAllTracks;
+	epv->getComment = impl_GNOME_Media_CDDBSlave2_getComment;
+	epv->setComment = impl_GNOME_Media_CDDBSlave2_setComment;
+	epv->getYear = impl_GNOME_Media_CDDBSlave2_getYear;
+	epv->setYear = impl_GNOME_Media_CDDBSlave2_setYear;
+	epv->getGenre = impl_GNOME_Media_CDDBSlave2_getGenre;
+	epv->setGenre = impl_GNOME_Media_CDDBSlave2_setGenre;
 }
 
 static void
@@ -905,6 +1438,10 @@ cddb_slave_init (CDDBSlave *cddb)
 
 	priv = g_new0 (CDDBSlavePrivate, 1);
 	cddb->priv = priv;
+
+	if (cddb_cache == NULL) {
+		cddb_cache = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+	}
 }
 
 BONOBO_TYPE_FUNC_FULL (CDDBSlave, GNOME_Media_CDDBSlave2, PARENT_TYPE, cddb_slave);
@@ -929,6 +1466,8 @@ BONOBO_TYPE_FUNC_FULL (CDDBSlave, GNOME_Media_CDDBSlave2, PARENT_TYPE, cddb_slav
 CDDBSlave *
 cddb_slave_new_full (const char *server,
 		     int port,
+		     const char *name,
+		     const char *hostname,
 		     BonoboEventSource *event_source)
 {
 	CDDBSlave *cddb;
@@ -951,6 +1490,9 @@ cddb_slave_new_full (const char *server,
 		priv->port = port;
 	}
 
+	priv->name = g_strdup (name);
+	priv->hostname = g_strdup (hostname);
+	
 	priv->event_source = event_source;
 	bonobo_object_add_interface (BONOBO_OBJECT (cddb),
 				     BONOBO_OBJECT (priv->event_source));
@@ -971,13 +1513,15 @@ cddb_slave_new_full (const char *server,
  */
 CDDBSlave *
 cddb_slave_new (const char *server,
-		int port)
+		int port,
+		const char *name,
+		const char *hostname)
 {
 	BonoboEventSource *event_source;
 
 	event_source = bonobo_event_source_new ();
 
-	return cddb_slave_new_full (server, port, event_source);
+	return cddb_slave_new_full (server, port, name, hostname, event_source);
 }
 
 /**
